<?php
/**
 * ownCloud
 *
 * @author Juan Pablo Villafañez Ramos <jvillafanez@owncloud.com>
 * @author Jesus Macias Portela <jesus@owncloud.com>
 * @copyright (C) 2014 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

/**
 * REQUIREMENTS:
 * - libsmbclient and libsmbclient-dev packages installed
 * - libsmbclient-php bindings installed and libsmbclient.so active for php
 * - HOME variable accessible for php (due to a bug in libsmbclient-php)
 */

namespace OCA\windows_network_drive\lib;

use \Icewind\Streams\CallbackWrapper;
use \Icewind\SMB\Exception\AlreadyExistsException;
use \Icewind\SMB\Exception\NotFoundException;
use \Icewind\SMB\Exception\Exception;
use \Icewind\SMB\NativeServer;
use \OC\Files\Stream\Dir;
use \OC\Files\Filesystem;

/**
 * FS class to connect to windows network drives
 */
class WND extends \OC\Files\Storage\Common {
    /** Enable debug */
    private static $enableLogs = FALSE;
    /** Password for the connection */
    private $password;
    /** User to connect */
    private $user;
    /** Host where the connection should be made */
    private $host;
    /** The folder that will act as root */
    private $root;
    /** The share name that we'll use */
    private $shareName;

    /** is sharing active for this FS? */
    private $isSharingActive = false;
    /** a cache for the dir call */
    private $dirCache = array();
    private static $tempFiles = array();
    /** a cache for attributes */
    private $xattrCache = array();

    /** Reference to a SMB server (check 3rdparty native server) */
    private $server;
    /** Reference to a SMB share (check 3rdparty native share) */
    private $share;


    /**
     * Construct the object. The required params are: "host", "user", "password" and "share".
     * "root" is optional, defaulting to '/'
     *
     * @param array $params an array with the required params
     */
    public function __construct($params) {
        // if user is null we convert username and password to '' to try anonymous login
        $params = $this->fixConstructParams($params);
        $this->isSharingActive = $params['oc_sharing'];
        if (isset($params['host']) && isset($params['share'])) {
            $params['share'] = trim($params['share'], '/');
            $this->shareName = $params['share'];
            $this->host = $params['host'];
            $this->user = $params['user'];
            $this->password = $params['password'];
            $this->root = $params['root'];
            $this->server = new NativeServer($this->host, $this->user, $this->password);
            $this->share = $this->server->getShare($params['share']);
        } else {
            throw new \Exception('invalid parameters in the wnd constructor');
        }
    }

    private function fixConstructParams($params) {
        $p = $params;
        unset($p['password']);
        $this->log('fixConstructParams: '.print_r($p, true), \OCP\Util::DEBUG);
        // if user is null we convert username and password to '' to try anonymous login
        $params['user'] = Utils::getDefault($params, 'user', '');
        $params['password'] = Utils::getDefault($params, 'password', '');

        //make sure the root end like "/something/" or "/something/more/"
        $params['root'] = Utils::getDefault($params, 'root', '');
        $params['root'] = '/' . trim($params['root'], '/') . '/';
        //fix special case
        if ($params['root'] === '//') {
            $params['root'] = '/';
        }

        $params['oc_sharing'] = (bool) Utils::getDefault($params, 'oc_sharing', false);
        return $params;
    }

    /**
     * Get the username trying to access to the windows drive. It might be different than the
     * current ownCloud user
     *
     * @return string the username
     */
    public function getUser() {
        $this->log('getUser: '.$this->user, \OCP\Util::DEBUG);
        return $this->user;
    }

    /**
     * Get the host we're trying to access
     *
     * @return string the host
     */
    public function getHost() {
        $this->log('getHost: '.$this->host, \OCP\Util::DEBUG);
        return $this->host;
    }

    /**
     * Get the sharename where we're trying to access
     *
     * @return string the share name
     */
    public function getShareName() {
        $this->log('getShareName: '.$this->shareName, \OCP\Util::DEBUG);
        return $this->shareName;
    }

    /**
     * Get the folder that will act as our root
     *
     * @return string the root folder
     */
    public function getRoot() {
        $this->log('getRoot: '.$this->root, \OCP\Util::DEBUG);
        return $this->root;
    }

    /**
     * Get the password we'll to to connect to the windows drive. Use it only for unittest
     *
     * @return string the password we'll use
     */
    public function getPassword() {
        $this->log('getPassword', \OCP\Util::DEBUG);
        return $this->password;
    }

    /**
     * Get the FS id (for ownCloud purposes)
     *
     * @return string the FS id
     */
    public function getId(){
        $id = 'wnd::' . $this->getUser() . '@' . $this->getHost() . '/' . $this->getShareName() . '/' . $this->getRoot();
        $this->log('getId id: '.$id, \OCP\Util::DEBUG);
        return $id;
    }

    private function dir($path) {
        $this->log('dir path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            if (!isset($this->dirCache[$path])) {
                $this->dirCache[$path] = array();
                $dirData = $this->share->dir($path);
                foreach ($dirData as $fileInfo) {
                    $this->dirCache[$path][$fileInfo->getName()] = $fileInfo;
                }
            }
            return $this->dirCache[$path];
        } catch (NotFoundException $e) {
            return array();
        } catch (\Exception $e){
            $this->log('function dir - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            throw $e;
        }
    }

    private function removeFromCache($path) {
        $this->log('removeFromCache path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        if (isset($this->dirCache[$path])) {
            unset($this->dirCache[$path]);
        }

        if ($this->isRootDir(dirname($path))) {
            // if we need to remove the root, wipe the cache
            $this->dirCache = array();
        } else {
            if (isset($this->dirCache[dirname($path)])) {
                unset($this->dirCache[dirname($path)]);
            }
        }
    }


    /**
     * Create a directory. Notice that all the parent directories should exists
     *
     * @param string $path the directory we want to create
     * @return bool true if the directory is created, false otherwise
     */
    public function mkdir($path) {
        $this->log('mkdir path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            if (!$this->file_exists($path)) {
                $this->removeFromCache($this->root . $path);
                $this->share->mkdir($this->root . $path);
                return true;
            } else {
                return false;
            }
        } catch (\Exception $e){
            $this->log('function mkdir - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        }
    }

    /**
     * Remove the directory and its contents
     *
     * @param string $path the FS path that we want to delete
     * @return bool true if deleted, false otherwise
     */
    public function rmdir($path) {
        $this->log('rmdir path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            $content = $this->dir($this->root . $path);
            foreach ($content as $name => $fileInfo) {
                if (!$fileInfo->isDirectory()) {
                    $this->unlink($path . '/' . $name);
                } else {
                    $this->rmdir($path . '/' . $name);
                }
            }
            $this->share->rmdir($this->root . $path);
            $this->removeFromCache($this->root . $path);
            return true;
        } catch (\Exception $e){
            $this->log('function rmdir - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        }
    }

    /**
     * Open the directory
     *
     * @param string $path FS path
     * @return resource a directory resource
     */
    public function opendir($path) {
        $this->log('opendir path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            $files = $this->dir($this->root . $path);
        } catch (\Exception $e){
            $this->log('function opendir - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            $files = array();
        }
        Dir::register('wnd:' . $path, array_keys($files));
        return opendir('fakedir://wnd:' . $path);
    }

    /**
     * Get the file type of the path
     *
     * @param string $path FS path
     * @return string the type of the path (usually 'dir' or 'file')
     */
    public function filetype($path) {
        $this->log('filetype path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        if ($this->isRootDir($path)) {
            return 'dir';
        }

        $stat = $this->stat($path);
        if ($stat) {
            return $stat['type'];
        } else {
            return false;
        }
    }

    /**
     * Check the the file defined by path exists
     *
     * @param string $path FS path
     * @return bool true if the file exists, false otherwise
     */
    public function file_exists($path) {
        $this->log('file_exists path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        if ($this->isRootDir($path) && $this->remoteIsShare()) {
            // refering to a share (not folder) -> return true
            return true;
        }

        try {
            $parent = dirname($this->root . $path);
            $files = $this->dir($parent);
            return isset($files[basename($this->root . $path)]);
        } catch (\Exception $e){
            $this->log('function file_exists - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        }
    }

    /**
     * Remove the path. If the path is a directory, this function falls back to the rmdir
     * function
     *
     * @param string $path FS path to be deleted
     * @return bool true if the path no longer exists, false otherwise
     */
    public function unlink($path) {
        $this->log('unlink path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            if ($this->is_dir($path)) {
                return $this->rmdir($path);
            } else {
                $this->removeFromCache($this->root . $path);
                return $this->share->del($this->root . $path);
            }
        } catch (\Exception $e){
            $this->log('function unlink - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        }
    }

    /**
     * Open a file resouce
     *
     * @param string $path file path to be opened
     * @param string $mode open the file in this mode
     * @return resource an opened resource to read or write
     */
    public function fopen($path, $mode) {
        $this->log('fopen path: '.$path.' mode: '.$mode, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        $fullPath = $this->buildPath($path);
        try {
            switch ($mode) {
                case 'r':
                case 'rb':
                    if (!$this->file_exists($path)) {
                        return false;
                    }
                    return $this->share->read($fullPath);
                case 'w':
                case 'wb':                    
                    $source = $this->share->write($fullPath);
                    return CallBackWrapper::wrap($source, null, null, function () use ($path) {
                        $this->removeFromCache($this->root . $path);
                    });
                case 'a':
                case 'ab':
                case 'r+':
                case 'w+':
                case 'wb+':
                case 'a+':
                case 'x':
                case 'x+':
                case 'c':
                case 'c+':
                    //emulate these
                    $tempManager = \OC::$server->getTempManager();
                    if (strrpos($path, '.') !== false) {
                        $ext = substr($path, strrpos($path, '.'));
                    } else {
                        $ext = '';
                    }
                    if ($this->file_exists($path)) {
                        if (!$this->isUpdatable($path)) {
                            return false;
                        }
                        if ($mode === 'w' or $mode === 'w+') {
                            $tmpFile = $tempManager->getTemporaryFile($ext);
                        } else {
                            $tmpFile = $this->getCachedFile($path);
                        }
                    } else {
                        if (!$this->isCreatable(dirname($path))) {
                            return false;
                        }
                        $tmpFile = $tempManager->getTemporaryFile($ext);
                    }
                    $source = fopen($tmpFile, $mode);
                    $share = $this->share;
                    return CallBackWrapper::wrap($source, null, null, function () use ($tmpFile, $fullPath, $path, $share) {
                        $this->removeFromCache($this->root . $path);
                        $share->put($tmpFile, $fullPath);
                        unlink($tmpFile);
                    });
            }
            return false;
        } catch (NotFoundException $e){
            $this->log('function fopen - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        } catch (Exception $e){
            $this->log('function fopen - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;          
        } catch (\Exception $e){
            $this->log('function fopen - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            throw $e;
        }
    }

    /**
     * "Touch" the path. This function will only create a empty file if not exists or update
     * the mtime to the latest if it exists. Setting the mtime is not supported
     *
     * @param string $path path to be "touched"
     * @param mixed $mtime null to be updated to the latest, if it's not null, it will be ignored
     * @return bool true if file exists and mtime updated
     */
    public function touch($path, $time = null) {
        $this->log('touch path: '.$path.' time: '.$time, \OCP\Util::DEBUG);
        if (!is_null($time)) {
            return false;
        }
        if (!$this->file_exists($path)) {
            try{
                $this->removeFromCache($this->root . $path);
                $fh = $this->share->write($this->buildPath($path));
            } catch (\Exception $e){
                $this->log('function touch - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
                throw $e;
            }
            fclose($fh);
            return true;
        }
        return false;
    }

    /**
     * Get the mimetype of the path or false if the path not exists
     *
     * @param string $path the target path
     * @return string|false the mimetype or false
     */
    public function getMimeType($path) {
        $this->log('getMimeType path: '.$path, \OCP\Util::DEBUG);
        if ($this->is_dir($path)) {
            return 'httpd/unix-directory';
        }
        if (!$this->file_exists($path)) {
            return false;
        }
        return \OC_Helper::getFileNameMimeType($path);
    }


    /**
     * Rename the files
     *
     * @param string $path1 the old name of the path
     * @param string $path2 the new name of the path
     * @return bool true if the rename is successful, false otherwise
     */
    public function rename($source, $target) {
        $this->log('rename source: '.$source.' target:'.$target, \OCP\Util::DEBUG);
        try {
            $result = $this->share->rename($this->root . $source, $this->root . $target);
            $this->removeFromCache($this->root . $source);
            $this->removeFromCache($this->root . $target);
        } catch (AlreadyExistsException $e) {
            $this->unlink($target);
            $result = $this->share->rename($this->root . $source, $this->root . $target);
            $this->removeFromCache($this->root . $source);
            $this->removeFromCache($this->root . $target);
            $this->log('function rename - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
        } catch (\Exception $e) {
            $this->log('function rename - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            $result = false;
        }
        return $result;
    }

    /**
     * perform the stat call over the file
     *
     * @param string $path the FS path
     * @param array|false an array with information or false in case of error
     */
    public function stat($path) {
        $this->log('stat path: '.$path, \OCP\Util::DEBUG);
        $path = trim($path, '/');
        try {
            if ($this->remoteIsShare() && $this->isRootDir($path)) { //mtime doesn't work for shares
                return array('mtime' => $this->shareMTime(), 'size' => 0, 'type' => 'dir');
            } else {
                $parent = dirname($this->root . $path);
                $files = $this->dir($parent);
                if (isset($files[basename($this->root . $path)])) {
                    $file = $files[basename($this->root . $path)];
                    return array('mtime' => $file->getMTime(),
                                'size' => $file->getSize(),
                                'type' => ($file->isDirectory()) ? 'dir' : 'file');
                } else {
                    return false;
                }
            }
        } catch (\Exception $e){
            $this->log('function stat - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return false;
        }
    }

    /**
     * Check the path has updated in the given time. The check is usually based on the mtime
     * of the path, so if the path's filemtime is greater than $time, the file has updated
     *
     * @param string $path the FS path
     * @param int $time timestamp to compare against
     * @throw \OCP\Files\StorageNotAvailableException
     * @return bool true if the file has updated, false otherwise
     */
    public function hasUpdated($path,$time) {
        $this->log('hasUpdated path: '.$path.' time: '.$time, \OCP\Util::DEBUG);
        $stat = $this->stat($path);
        if ($stat === false) {
            \OCP\Util::writeLog('wnd', 'hasUpdated failed -> storage not available', \OCP\Util::ERROR);
            throw new \OCP\Files\StorageNotAvailableException();
            // if no stat consider a the file has updated to force a rescan
            //return true;
        } else {
            $currentMtime = $stat['mtime'];
            return $currentMtime > $time;
        }
    }

    /**
     * get the best guess for the modification time of the share
     *
     * @param string $path the target path
     * @return array the first element of the array is the calculated mtime for the folder, and
     * the second is the list of readable files (the list contains the filename as key and the
     * mtime for the file as value)
     */
    private function shareMTime() {
        $files = $this->dir($this->root);
        $lastMtime = 0;
        foreach ($files as $file) {
            if ($file->getMTime() > $lastMtime) {
                $lastMtime = $file->getMTime();
            }
        }
        $this->log('shareMTime: '.$lastMtime, \OCP\Util::DEBUG);
        return $lastMtime;
    }

    /**
     * Check if the path is our root dir (not the smb one)
     *
     * @param string $path the path
     * @return bool true if it's root, false if not
     */
    private function isRootDir($path) {
        $result = ($path == '' || $path == '/' || $path == '.');
        $this->log('isRootDir path:'.$path.' result: '.(string)$result, \OCP\Util::DEBUG);
        return $result;
    }

    /**
     * Check if our root points to a smb share
     *
     * @return bool true if our root points to a share false otherwise
     */
    private function remoteIsShare() {
        $result = $this->getShareName() && (!$this->getRoot() || $this->getRoot() === '/');
        $this->log('remoteIsShare:'.(string)$result, \OCP\Util::DEBUG);
        return $result;
    }

    /**
     * Check if our root points to a smb path
     *
     * @return bool true if our root points to a path false otherwise
     */
    private function remoteIsFolder() {
        $result = $this->getShareName() && ($this->getRoot() && $this->getRoot() !== '/');
        $this->log('remoteIsFolder: '.(string)$result, \OCP\Util::DEBUG);
        return $result;
    }

    /**
     * Check where our root points to remotely. It can be a smb share ("share") or one of the
     * inner folder ("folder")
     *
     * @return string 'share' or 'folder'
     */
    public function remoteType() {
        $this->log('remoteType', \OCP\Util::DEBUG);
        if ($this->remoteIsShare()) return 'share';
        if ($this->remoteIsFolder()) return 'folder';
    }

    /**
     * Check if the path is readable
     *
     * @param string $path the path
     * @return bool true if it's readable, false if not
     */
    public function isReadable($path) {
        $fileInfoOperation = (($info = $this->getFileInfoOperation($path, 'isReadOnly')) === null) ? false : !$info;
        $this->log('isCreatable path: '.$path.' - '.$fileInfoOperation, \OCP\Util::DEBUG);
        if ($this->isRootDir($path)) {
            return true;
        }
        $this->log('isReadable path: '.(string)$fileInfoOperation, \OCP\Util::DEBUG);
        return $fileInfoOperation;
    }

    /**
     * Check if the path is updatable
     *
     * @param string $path the path
     * @return bool true if it's updatable, false if not
     */
    public function isUpdatable($path) {
        $fileInfoOperation = (($info = $this->getFileInfoOperation($path, 'isReadOnly')) === null) ? false : !$info;
        $this->log('isUpdatable path: '.$path.' - '.$fileInfoOperation, \OCP\Util::DEBUG);
        $this->log('isUpdatable path: '.(string)$fileInfoOperation, \OCP\Util::DEBUG);
        return $fileInfoOperation;
    }

    /**
     * Check if the path is creatable
     *
     * @param string $path the path
     * @return bool true if it's creatable, false if not
     */
    public function isCreatable($path) {
        $fileInfoOperation = (($info = $this->getFileInfoOperation($path, 'isReadOnly')) === null) ? false : !$info;
        $this->log('isCreatable path: '.$path.' - '.$fileInfoOperation, \OCP\Util::DEBUG);
        if (!$this->isRootDir($path)) {
            // if the path isn't root, use the default behaviour
            $creatable = parent::isCreatable($path);
            $this->log('isCreatable path: '.(string)$creatable, \OCP\Util::DEBUG);
            return $creatable;
        }
        // if path is root check that the folder isn't read-only
        $this->log('isCreatable path: '.(string)$fileInfoOperation, \OCP\Util::DEBUG);
        return $fileInfoOperation;
    }

    private function getFileInfoOperation($path, $operation) {
        $this->log('getFileInfoOperation path: '.$path.' operation: '.$operation, \OCP\Util::DEBUG);
        if (isset($this->xattrCache[$path])) {
             $nativeFI = $this->xattrCache[$path];
        } else {
             $realPath = $this->root . $path;
             $nativeFI = new \Icewind\SMB\NativeFileInfo($this->share, $realPath, basename($realPath));
             $this->xattrCache[$path] = $nativeFI;
        }

        try {
            return call_user_func(array($nativeFI, $operation));
        } catch (\Exception $e){
            $this->log('function getFileInfoOperation - code: '.$e->getCode().' message: '.$e->getMessage().' trace: '.$e->getTraceAsString(), \OCP\Util::DEBUG);
            return null;
        }
    }

    /**
     * Check if the path is sharable. Currently sharing is disabled so it will always
     * return false
     *
     * @param string $path the path
     * @return false
     */
    public function isSharable($path) {
        $this->log('isSharable path: '.$path, \OCP\Util::DEBUG);
        return $this->isSharingActive;
    }

    /**
     * @param string $path
     * @return string
     */
    protected function buildPath($path) {
        $this->log('buildPath path: '.$path, \OCP\Util::DEBUG);
        return Filesystem::normalizePath($this->root . '/' . $path);
    }


    /**
     * @param string $message
     */
    private function log($message, $level, $from='wnd') {
        if(\OC::$server->getConfig()->getSystemValue('wnd.logging.enable', false) === true){
            \OCP\Util::writeLog($from, $message, $level);
        }
    }

    /**
     * A simple function to test connectivity. This functions might throw exceptions
     */
    public function test() {
        $this->share->dir(trim($this->root, '/'));
        return true;
    }
}
